/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#ifndef _WIN32
#define GLX_GLXEXT_PROTOTYPES
#include <GL/glx.h>
#endif

#ifdef _WIN32
#define qglGetProcAddress wglGetProcAddress
#else
#define glXGetProcAddress glXGetProcAddressARB
#define qglGetProcAddress(x) glXGetProcAddress((const GLubyte *)(x))
#endif

#include "quakedef.h"

cvar_t		gl_picmip = {"gl_picmip", "0"};
cvar_t		gl_texturequality = {"gl_texturequality", "1"};
cvar_t		r_conalpha = {"r_conalpha", "1"};

extern vrect_t		scr_vrect;
extern cvar_t crosshair, cl_crossx, cl_crossy;

byte		*draw_chars;				// 8*8 graphic characters
qpic_t		*draw_disc;
qpic_t		*draw_backtile;

gltexture_t *notexture;
gltexture_t *nulltexture;
gltexture_t	*char_texture; 

typedef struct
{
	gltexture_t	*gltexture;
	float	sl, tl, sh, th;
} glpic_t;

int		indexed_bytes = 1;
int		rgba_bytes = 4;

int		gl_solid_format = GL_RGB; // was 3
int		gl_alpha_format = GL_RGBA; // was 4

#define MAXGLMODES 6

typedef struct
{
	int	magfilter;
	int minfilter;
	char *name;
} glmode_t;

glmode_t modes[MAXGLMODES] = {
	{GL_NEAREST, GL_NEAREST,				"GL_NEAREST"},
	{GL_NEAREST, GL_NEAREST_MIPMAP_NEAREST,	"GL_NEAREST_MIPMAP_NEAREST"},
	{GL_NEAREST, GL_NEAREST_MIPMAP_LINEAR,	"GL_NEAREST_MIPMAP_LINEAR"},
	{GL_LINEAR,  GL_LINEAR,					"GL_LINEAR"},
	{GL_LINEAR,  GL_LINEAR_MIPMAP_NEAREST,	"GL_LINEAR_MIPMAP_NEAREST"},
	{GL_LINEAR,  GL_LINEAR_MIPMAP_LINEAR,	"GL_LINEAR_MIPMAP_LINEAR"},
};

int		gl_texturemode = 3; // linear
int		gl_filter_min = GL_LINEAR; // was GL_NEAREST
int		gl_filter_mag = GL_LINEAR;
float 	gl_textureanisotropy = 1;

int 	warpimagesize;

#define	MAX_GLTEXTURES	2048 // was 1024
gltexture_t	*active_gltextures, *free_gltextures;
int			numgltextures;

int currenttexture = -1; // to avoid unnecessary texture sets
GLenum TEXTURE0, TEXTURE1;
qboolean mtexenabled = false;

// move this to gl_draw.c
unsigned int d_8to24table[256]; // for GL renderer
unsigned int d_8to24table_fbright[256];
unsigned int d_8to24table_nobright[256];
unsigned int d_8to24table_conchars[256];

// move to gl_draw.c
const char *gl_vendor;
const char *gl_renderer;
const char *gl_version;
const char *gl_extensions;
qboolean fullsbardraw = false;

// move to gl_draw.c
qboolean gl_mtexable = false;
qboolean gl_texture_env_combine = false;
qboolean gl_texture_env_add = false;
float gl_maxanisotropy;
int gl_hardwaremaxsize;
int gl_texturemaxsize;

/*
================================================

			OpenGL Stuff

================================================
*/

// move to gl_draw.c
/*
===============
GL_CheckExtensions
===============
*/
void GL_CheckExtensions (void) 
{
	qboolean ARB = false;
	qboolean EXTcombine, ARBcombine;
	qboolean EXTadd, ARBadd;
	qboolean anisotropy;
	int units;

	// Multitexture
	ARB = strstr (gl_extensions, "GL_ARB_multitexture ") != NULL;

	if (COM_CheckParm("-nomtex"))
	{
		Con_Warning ("Multitexture disabled at command line\n");
	}
	else if (ARB)
	{
		Con_Printf ("GL_ARB_multitexture extensions found\n");

		qglMultiTexCoord2f = (void *) qglGetProcAddress ("glMultiTexCoord2fARB");
		qglActiveTexture = (void *) qglGetProcAddress ("glActiveTextureARB");

		TEXTURE0 = GL_TEXTURE0_ARB;
		TEXTURE1 = GL_TEXTURE1_ARB;

		gl_mtexable = true;
	}
	else
	{
		Con_Warning ("Multitexture not supported\n");
	}

	// Check how many texture units there actually are
	if (ARB)
	{
		glGetIntegerv (GL_MAX_TEXTURE_UNITS_ARB, &units);
		Con_Printf ("%i TMUs found\n", units);
	}

	// poll max size from hardware
	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &gl_hardwaremaxsize);
	Con_Printf ("textures maximum size is %i\n", gl_hardwaremaxsize);
	gl_texturemaxsize = gl_hardwaremaxsize; // by default we sets maxsize as hardware maxsize

	// Texture Combine Environment mode
	ARBcombine = strstr (gl_extensions, "GL_ARB_texture_env_combine ") != NULL;
	EXTcombine = strstr (gl_extensions, "GL_EXT_texture_env_combine ") != NULL;

	if (COM_CheckParm("-nocombine"))
	{
		Con_Warning ("Texture Combine Environment disabled at command line\n");
	}
	else if (ARBcombine || EXTcombine)
	{
		Con_Printf ("GL_%s_texture_env_combine extensions found\n", ARBcombine ? "ARB" : "EXT");
		gl_texture_env_combine = true;
	}
	else
	{
		Con_Warning ("Texture Combine Environment not supported\n");
	}

	// Texture Add Environment mode
	ARBadd = strstr (gl_extensions, "GL_ARB_texture_env_add ") != NULL;
	EXTadd = strstr (gl_extensions, "GL_EXT_texture_env_add ") != NULL;

	if (COM_CheckParm("-noadd"))
	{
		Con_Warning ("Texture Add Environment disabled at command line\n");
	}
	else if (ARBadd || EXTadd)
	{
		Con_Printf ("GL_%s_texture_env_add extensions found\n", ARBadd ? "ARB" : "EXT");
		gl_texture_env_add = true;
	}
	else
	{
		Con_Warning ("Texture Add Environment not supported\n");
	}

	// Anisotropic Filtering
	anisotropy = strstr (gl_extensions, "GL_EXT_texture_filter_anisotropic") != NULL;

	if (anisotropy)
	{
		float test1,test2;
		int tex;

		// test to make sure we really have control over it
		// 1.0 and 2.0 should always be legal values
		glGenTextures (1, (GLuint *)&tex);
		glBindTexture (GL_TEXTURE_2D, tex);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 1.0f);
		glGetTexParameterfv (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, &test1);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 2.0f);
		glGetTexParameterfv (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, &test2);
		glDeleteTextures (1, (const GLuint *)&tex);

		if (test1 == 1 && test2 == 2)
			Con_Printf ("GL_EXT_texture_filter_anisotropic extensions found\n");
		else
			Con_Warning ("Anisotropic Filtering locked by driver. Current driver setting is %f\n", test1);

		//get max value either way, so the menu and stuff know it
		glGetFloatv (GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &gl_maxanisotropy);
	}
	else
	{
		Con_Warning ("Anisotropic Filtering not supported\n");
	}
}

/*
===============
GL_MakeExtensionsList
===============
*/
char *GL_MakeExtensionsList (const char *in)
{
	char *copy, *token, *out;
	int i, count;

	//each space will be replaced by 4 chars, so count the spaces before we malloc
	for (i = 0, count = 1; i < (int)strlen(in); i++)
		if (in[i] == ' ')
			count++;
	out = Z_Malloc (strlen(in) + count*3 + 1); //usually about 1-2k
	out[0] = 0;

	copy = Z_Malloc(strlen(in) + 1);
	strcpy(copy, in);

	for (token = strtok(copy, " "); token; token = strtok(NULL, " "))
	{
		strcat(out, "\n   ");
		strcat(out, token);
	}

	Z_Free (copy);
	return out;
}

/*
===============
GL_Info_f
===============
*/
void GL_Info_f (void)
{
	static char *gl_extensions_nice = NULL;

	if (!gl_extensions_nice)
		gl_extensions_nice = GL_MakeExtensionsList (gl_extensions);

	Con_SafePrintf ("GL_VENDOR: %s\n", gl_vendor);
	Con_SafePrintf ("GL_RENDERER: %s\n", gl_renderer);
	Con_SafePrintf ("GL_VERSION: %s\n", gl_version);
	Con_SafePrintf ("GL_EXTENSIONS: %s\n", gl_extensions_nice);
}

/*
===============
GL_Init
===============
*/
void GL_Init (void)
{
	// gl_info
	gl_vendor = (const char *)glGetString (GL_VENDOR);
//	Con_SafePrintf ("GL_VENDOR: %s\n", gl_vendor);
	gl_renderer = (const char *)glGetString (GL_RENDERER);
//	Con_SafePrintf ("GL_RENDERER: %s\n", gl_renderer);
	gl_version = (const char *)glGetString (GL_VERSION);
//	Con_SafePrintf ("GL_VERSION: %s\n", gl_version);
	gl_extensions = (const char *)glGetString (GL_EXTENSIONS);
//	Con_SafePrintf ("GL_EXTENSIONS: %s\n", gl_extensions);

	if (strncasecmp(gl_renderer,"PowerVR",7)==0)
		fullsbardraw = true;

	GL_CheckExtensions ();

	Cmd_AddCommand ("gl_info", GL_Info_f);

	glClearColor (0.15,0.15,0.15,0); // originally was 1,0,0,0

	glCullFace(GL_BACK); // glquake used CCW with backwards culling -- let's do it right
	glFrontFace(GL_CW); // glquake used CCW with backwards culling -- let's do it right

	glEnable(GL_TEXTURE_2D);

	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.666);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel (GL_FLAT);

	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE); 

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // was GL_NEAREST
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // was GL_NEAREST
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glDepthRange (0, 1); // moved here because gl_ztrick is gone.
	glDepthFunc (GL_LEQUAL); // moved here because gl_ztrick is gone.
}

/*
===============
GL_SetPalette
===============
*/
void GL_SetPalette (byte *palette)
{
// TODO: move this code into func GL_SetPalette
// and move it to gl_draw.c
	byte mask[] = {255,255,255,0};
	byte black[] = {0,0,0,255};
	byte *pal, *src, *dst;
	int i;

	pal = palette;

	//standard palette, 255 is transparent
	dst = (byte *)d_8to24table;
	src = pal;
	for (i=0; i<256; i++)
	{
		*dst++ = *src++;
		*dst++ = *src++;
		*dst++ = *src++;
		*dst++ = 255;
	}
	d_8to24table[255] &= *mask; // fix gcc warnings

	//fullbright palette, 0-223 are black (for additive blending)
	src = pal + 224*3;
	dst = (byte *)(d_8to24table_fbright) + 224*4;
	for (i=224; i<256; i++)
	{
		*dst++ = *src++;
		*dst++ = *src++;
		*dst++ = *src++;
		*dst++ = 255;
	}
	for (i=0; i<224; i++)
		d_8to24table_fbright[i] = *black; // fix gcc warnings

	//nobright palette, 224-255 are black (for additive blending)
	dst = (byte *)d_8to24table_nobright;
	src = pal;
	for (i=0; i<256; i++)
	{
		*dst++ = *src++;
		*dst++ = *src++;
		*dst++ = *src++;
		*dst++ = 255;
	}
	for (i=224; i<256; i++)
		d_8to24table_nobright[i] = *black; // fix gcc warnings

	//conchars palette, 0 and 255 are transparent
	memcpy(d_8to24table_conchars, d_8to24table, 256*4);
	d_8to24table_conchars[0] &= *mask; // fix gcc warnings
}

/*
================
GL_Bind
================
*/
void GL_Bind (gltexture_t *texture)
{
	if (!texture)
		texture = nulltexture;

	if (texture->texnum != currenttexture)
	{
		currenttexture = texture->texnum;
		glBindTexture (GL_TEXTURE_2D, currenttexture);
	}
}

/*
================
GL_SelectTexture
================
*/
void GL_SelectTexture (GLenum target)
{
	static GLenum currenttarget;
	static int ct0, ct1;

	if (target == currenttarget)
		return;

	qglActiveTexture (target);

	if (target == TEXTURE0)
	{
		ct1 = currenttexture;
		currenttexture = ct0;
	}
	else //target == TEXTURE1
	{
		ct0 = currenttexture;
		currenttexture = ct1;
	}

	currenttarget = target;
}

/*
================
GL_DisableMultitexture -- selects texture unit 0
================
*/
void GL_DisableMultitexture (void) 
{
	if (mtexenabled) 
	{
		glDisable(GL_TEXTURE_2D);
		GL_SelectTexture(TEXTURE0);
		mtexenabled = false;
	}
}

/*
================
GL_EnableMultitexture -- selects texture unit 1
================
*/
void GL_EnableMultitexture (void) 
{
	if (gl_mtexable) 
	{
		GL_SelectTexture(TEXTURE1);
		glEnable(GL_TEXTURE_2D);
		mtexenabled = true;
	}
}


/*
=============================================================================

  scrap allocation

  Allocate all the little status bar obejcts into a single texture
  to crutch up stupid hardware / drivers

=============================================================================
*/

#define	MAX_SCRAPS		2
#define	BLOCK_WIDTH		256
#define	BLOCK_HEIGHT	256

int			scrap_allocated[MAX_SCRAPS][BLOCK_WIDTH];
byte		scrap_texels[MAX_SCRAPS][BLOCK_WIDTH*BLOCK_HEIGHT];
qboolean	scrap_dirty;
gltexture_t	*scrap_textures[MAX_SCRAPS]; // changed to array

// returns a texture number and the position inside it
int Scrap_AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		texnum;

	for (texnum=0 ; texnum<MAX_SCRAPS ; texnum++)
	{
		best = BLOCK_HEIGHT;

		for (i=0 ; i<BLOCK_WIDTH-w ; i++)
		{
			best2 = 0;

			for (j=0 ; j<w ; j++)
			{
				if (scrap_allocated[texnum][i+j] >= best)
					break;
				if (scrap_allocated[texnum][i+j] > best2)
					best2 = scrap_allocated[texnum][i+j];
			}
			if (j == w)
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > BLOCK_HEIGHT)
			continue;

		for (i=0 ; i<w ; i++)
			scrap_allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	Sys_Error ("Scrap_AllocBlock: full");
	return 0; // make compiler happy
}

void Scrap_Upload (void)
{
	int		i;
	char name[64];

	for (i = 0; i < MAX_SCRAPS; i++) 
	{
		sprintf (name, "scrap%d", i);
		scrap_textures[i] = GL_LoadTexture (NULL, name, BLOCK_WIDTH, BLOCK_HEIGHT, SRC_INDEXED, scrap_texels[i], TEXPREF_ALPHA | TEXPREF_OVERWRITE | TEXPREF_NOPICMIP);
	}
	scrap_dirty = false;
}


//=============================================================================
/* Support Routines */

typedef struct cachepic_s
{
	char		name[MAX_QPATH];
	qpic_t		pic;
	byte		padding[32];	// for appended glpic
} cachepic_t;

#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

qpic_t *Draw_PicFromWad (char *name)
{
	qpic_t	*p;
	glpic_t	*gl;

	p = W_GetLumpName (name);

	// Sanity ...
	if (p->width & 0xC0000000 || p->height & 0xC0000000)
		Sys_Error ("Draw_PicFromWad: invalid dimensions (%dx%d) for '%s'", p->width, p->height, name);

	// Check if pic is big enough for the gl struct
	if (p->width * p->height < sizeof(glpic_t))
	{
		qpic_t		*p2 = malloc (sizeof(qpic_t) + sizeof(glpic_t)); // This leaks, kludge
		static qboolean WarnPrinted;

		if (!WarnPrinted)
		{
			WarnPrinted = true;

			Con_Warning ("Draw_PicFromWad: Image size insufficient (%d, normal min = %d) in '%s'\n", p->width * p->height, sizeof(glpic_t), name);
		}

		memset (p2, 0, sizeof(qpic_t) + sizeof(glpic_t));
		memcpy (p2, p, sizeof(qpic_t));
		p = p2;
	}

	gl = (glpic_t *)p->data;

	// load little ones into the scrap
	if (p->width < 64 && p->height < 64)
	{
		int		x, y;
		int		i, j, k;
		int		texnum;

		texnum = Scrap_AllocBlock (p->width, p->height, &x, &y);
		scrap_dirty = true;
		k = 0;
		for (i=0 ; i<p->height ; i++)
			for (j=0 ; j<p->width ; j++, k++)
				scrap_texels[texnum][(y+i)*BLOCK_WIDTH + x + j] = p->data[k];

		gl->gltexture = scrap_textures[texnum]; // changed to an array
		// no longer go from 0.01 to 0.99
		gl->sl = x/(float)BLOCK_WIDTH;
		gl->sh = (x+p->width)/(float)BLOCK_WIDTH;
		gl->tl = y/(float)BLOCK_WIDTH;
		gl->th = (y+p->height)/(float)BLOCK_WIDTH;
	}
	else
	{
		gl->gltexture = GL_LoadTexture (NULL, name, p->width, p->height, SRC_INDEXED, p->data, TEXPREF_ALPHA | TEXPREF_PAD | TEXPREF_NOPICMIP);
		gl->sl = 0;
		gl->sh = 1;
		gl->tl = 0;
		gl->th = 1;
	}
	return p;
}


/*
================
Draw_CachePic
================
*/
qpic_t	*Draw_CachePic (char *path)
{
	cachepic_t	*pic;
	int			i;
	qpic_t		*dat;
//	glpic_t		*gl;
	glpic_t		gl;

	for (pic=menu_cachepics, i=0 ; i<menu_numcachepics ; pic++, i++)
		if (!strcmp (path, pic->name))
			return &pic->pic;

	if (menu_numcachepics == MAX_CACHED_PICS)
		Sys_Error ("menu_numcachepics == MAX_CACHED_PICS (%d)", MAX_CACHED_PICS);
	menu_numcachepics++;
	strcpy (pic->name, path);

//
// load the pic from disk
//
	dat = (qpic_t *)COM_LoadTempFile (path);
	if (!dat)
		Sys_Error ("Draw_CachePic: failed to load %s", path);
	SwapPic (dat);

	// HACK HACK HACK --- we need to keep the bytes for
	// the translatable player picture just for the menu
	// configuration dialog
	if (!strcmp (path, "gfx/menuplyr.lmp"))
		memcpy (menuplyr_pixels, dat->data, dat->width*dat->height);

	pic->pic.width = dat->width;
	pic->pic.height = dat->height;

	// fix gcc warnings
	gl.gltexture = GL_LoadTexture (NULL, path, dat->width, dat->height, SRC_INDEXED, dat->data, TEXPREF_ALPHA | TEXPREF_PAD | TEXPREF_NOPICMIP);
	gl.sl = 0;
	gl.sh = 1;
	gl.tl = 0;
	gl.th = 1;
	memcpy (pic->pic.data, &gl, sizeof(glpic_t));

	return &pic->pic;
}

// unused function
void Draw_CharToConback (int num, byte *dest, int width)
{
	int		row, col;
	byte	*source;
	int		drawline;
	int		x;

	row = num>>4;
	col = num&15;
	source = draw_chars + (row<<10) + (col<<3);

	drawline = 8;

	while (drawline--)
	{
		for (x=0 ; x<8 ; x++)
			if (source[x] != 255)
				dest[x] = 0x60 + source[x];
		source += 128;
		dest += width;
	}

}

/*
===============
GL_TextureMaxSize_f
===============
*/
void GL_TextureMaxSize_f (void)
{
	if (Cmd_Argc() == 1)
	{
		Con_Printf ("gl_texturemaxsize is %d, hardware limit is %d\n", gl_texturemaxsize, gl_hardwaremaxsize);
		return;
	}

	gl_texturemaxsize = CLAMP(2, atoi (Cmd_Argv(1)), gl_hardwaremaxsize);
}

/*
===============
GL_SetFilterModes
===============
*/
void GL_SetFilterModes (gltexture_t *glt)
{
	GL_Bind (glt);

	if (glt->flags & TEXPREF_NEAREST)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	}
	else if (glt->flags & TEXPREF_LINEAR)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}
	else if (glt->flags & TEXPREF_MIPMAP)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, modes[gl_texturemode].magfilter);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, modes[gl_texturemode].minfilter);
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, modes[gl_texturemode].magfilter);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, modes[gl_texturemode].magfilter);
	}

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, gl_textureanisotropy);
}

/*
===============
GL_TextureMode_f
===============
*/
void GL_TextureMode_f (void)
{
	int i;
	char *arg;
	gltexture_t *glt;

	if (Cmd_Argc() == 1)
	{
		for (i=0 ; i< MAXGLMODES ; i++)
		{
			if (gl_filter_min == modes[i].minfilter)
			{
				Con_Printf ("gl_texturemode is %s (%d)\n", modes[i].name, i + 1);
				return;
			}
		}
	}

	for (i=0 ; i< MAXGLMODES ; i++)
	{
		arg = Cmd_Argv(1);
		if (!strcasecmp (modes[i].name, arg ) || (isdigit(*arg) && atoi(arg) - 1 == i))
			break;
	}
	if (i == MAXGLMODES)
	{
		Con_Printf ("bad filter name, available are:\n");
		for (i=0 ; i< MAXGLMODES ; i++)
			Con_Printf ("%s (%d)\n", modes[i].name, i + 1);
		return;
	}

	gl_texturemode = i;
	gl_filter_min = modes[gl_texturemode].minfilter;
	gl_filter_mag = modes[gl_texturemode].magfilter;

	// change all the existing texture objects
	for (glt=active_gltextures; glt; glt=glt->next)
		GL_SetFilterModes (glt);

	Sbar_Changed (); //sbar graphics need to be redrawn with new filter mode
}

/*
===============
GL_TextureAnisotropy_f
===============
*/
void GL_TextureAnisotropy_f (void)
{
	gltexture_t	*glt;

	if (Cmd_Argc() == 1)
	{
		Con_Printf ("gl_textureanisotropy is %f, hardware limit is %f\n", gl_textureanisotropy, gl_maxanisotropy);
		return;
	}

	gl_textureanisotropy = CLAMP(1.0f, atof (Cmd_Argv(1)), gl_maxanisotropy);

	for (glt=active_gltextures; glt; glt=glt->next)
		GL_SetFilterModes (glt);
}


/*
===============
Load_CharSet
===============
*/
static void Load_CharSet (void)
{
	int  i;
	byte *dest, *src;

	// load the console background and the charset
	// by hand, because we need to write the version
	// string into the background before turning
	// it into a texture
	draw_chars = W_GetLumpName ("conchars");
	for (i=0 ; i<256*64 ; i++)
		if (draw_chars[i] == 0)
			draw_chars[i] = 255;	// proper transparent color

	// Expand charset texture with blank lines in between to avoid in-line distortion
	dest = malloc (128 * 256);
	memset (dest, 0, 128 * 256);
	src = draw_chars;

	for (i = 0; i < 16; ++i)
		memcpy (&dest[8 * 128 * 2 * i], &src[8 * 128 * i], 8 * 128); // Copy each line

	// now turn them into textures
	char_texture = GL_LoadTexture (NULL, "conchars", 128, 256, SRC_INDEXED, dest, TEXPREF_ALPHA | TEXPREF_NEAREST | TEXPREF_NOPICMIP | TEXPREF_CONCHARS);

	free (dest);
}


/*
=============
CalcWarpImageSize
=============
*/
void CalcWarpImageSize (void)
{
	//
	// find the new correct size
	//
	warpimagesize = gl_texturemaxsize;

	while (warpimagesize > (int)vid.width)
		warpimagesize >>= 1;
	while (warpimagesize > (int)vid.height)
		warpimagesize >>= 1;
}

/*
===============
Draw_Init
===============
*/
void Draw_Init (void)
{
	int i;
	static byte notexture_data[16] = {159,91,83,255,0,0,0,255,0,0,0,255,159,91,83,255}; //black and pink checker
	static byte nulltexture_data[16] = {127,191,255,255,0,0,0,255,0,0,0,255,127,191,255,255}; //black and blue checker

	// init texture list
	free_gltextures = (gltexture_t *) Hunk_AllocName (MAX_GLTEXTURES * sizeof(gltexture_t), "gltextures");
	active_gltextures = NULL;
	for (i=0; i<MAX_GLTEXTURES-1; i++)
		free_gltextures[i].next = &free_gltextures[i+1];
	free_gltextures[i].next = NULL;
	numgltextures = 0;

	Cvar_RegisterVariable (&gl_picmip);
	Cvar_RegisterVariable (&gl_texturequality);
	Cvar_RegisterVariable (&r_conalpha);

	Cmd_AddCommand ("gl_texturemaxsize", &GL_TextureMaxSize_f);
	Cmd_AddCommand ("gl_texturemode", &GL_TextureMode_f);
	Cmd_AddCommand ("gl_textureanisotropy", &GL_TextureAnisotropy_f);

	// load notexture images
	notexture = GL_LoadTexture (NULL, "notexture", 2, 2, SRC_RGBA, notexture_data, TEXPREF_NEAREST | TEXPREF_PERSIST | TEXPREF_NOPICMIP);
	nulltexture = GL_LoadTexture (NULL, "nulltexture", 2, 2, SRC_RGBA, nulltexture_data, TEXPREF_NEAREST | TEXPREF_PERSIST | TEXPREF_NOPICMIP);

	// have to assign these here
	r_notexture_mip->gltexture = r_notexture_mip2->gltexture = notexture;

	// clear scrap and allocate gltextures
	memset(&scrap_allocated, 0, sizeof(scrap_allocated));
	memset(&scrap_texels, 255, sizeof(scrap_texels));
	Scrap_Upload (); //creates 2 empty textures

	Load_CharSet ();

	//
	// get the other pics we need
	//
	draw_disc = Draw_PicFromWad ("disc");
	draw_backtile = Draw_PicFromWad ("backtile");

	//set safe size for warpimages
	warpimagesize = 0;
	CalcWarpImageSize ();
}

/*
===============
Draw_Crosshair
===============
*/
void Draw_Crosshair (void)
{
	if (!crosshair.value) 
		return;

	Draw_Character (scr_vrect.x + scr_vrect.width/2 + cl_crossx.value, scr_vrect.y + scr_vrect.height/2 + cl_crossy.value, '+');
}

/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
static qboolean IsValid (int y, int num)
{
	if ((num & 127) == 32)
		return false; // space

	if (y <= -8)
		return false; // totally off screen

	return true;
}

static void Character (int x, int y, int num)
{
	int	row, col;
	float	frow, fcol, size, offset;

	num &= 255;

	row = num>>4;
	col = num&15;

	frow = row*0.0625;
	fcol = col*0.0625;
	size = 0.0625;
//	offset = 0.002; // slight offset to avoid in-between lines distortion
	offset = 0.03125; // offset to match expanded charset texture

	glTexCoord2f (fcol, frow);
	glVertex2f (x, y);
	glTexCoord2f (fcol + size, frow);
	glVertex2f (x+8, y);
	glTexCoord2f (fcol + size, frow + size - offset);
	glVertex2f (x+8, y+8);
	glTexCoord2f (fcol, frow + size - offset);
	glVertex2f (x, y+8);
}

void Draw_Character (int x, int y, int num)
{
	if (!IsValid (y, num))
		return;

	GL_Bind (char_texture);
	glBegin (GL_QUADS);

	Character (x, y, num);

	glEnd ();
}

/*
================
Draw_String
================
*/
void Draw_String (int x, int y, char *str)
{
	GL_Bind (char_texture);
	glBegin (GL_QUADS);

	while (*str)
	{
		if (IsValid (y, *str))
			Character (x, y, *str);

		str++;
		x += 8;
	}

	glEnd ();
}


/*
=============
Draw_AlphaPic
=============
*/
void Draw_AlphaPic (int x, int y, qpic_t *pic, float alpha)
{
	glpic_t			*gl;

	if (scrap_dirty)
		Scrap_Upload ();

	gl = (glpic_t *)pic->data;

	glDisable(GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glColor4f (1,1,1,alpha);
	GL_Bind (gl->gltexture);
	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x+pic->width, y);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x+pic->width, y+pic->height);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y+pic->height);
	glEnd ();
	glColor4f (1,1,1,1);
	glEnable(GL_ALPHA_TEST);
	glDisable (GL_BLEND);
}


/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
	glpic_t			*gl;

	if (scrap_dirty)
		Scrap_Upload ();

	gl = (glpic_t *)pic->data;

	glColor4f (1,1,1,1);
	GL_Bind (gl->gltexture);
	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x+pic->width, y);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x+pic->width, y+pic->height);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y+pic->height);
	glEnd ();
}


/*
=============
Draw_SubPic
=============
*/
void Draw_SubPic (int x, int y, qpic_t *pic, int srcx, int srcy, int width, int height)
{
	float			newsl, newtl, newsh, newth, oldglwidth, oldglheight; 
	glpic_t			*gl;

	if (scrap_dirty)
		Scrap_Upload ();

	gl = (glpic_t *)pic->data;

	oldglwidth = gl->sh - gl->sl;
	oldglheight = gl->th - gl->tl;

	newsl = gl->sl + (srcx * oldglwidth) / pic->width;
	newsh = newsl + (width * oldglwidth) / pic->width;

	newtl = gl->tl + (srcy * oldglheight) / pic->height;
	newth = newtl + (height * oldglheight) / pic->height; 

	glColor4f (1,1,1,1);
	GL_Bind (gl->gltexture);
	glBegin (GL_QUADS);
	glTexCoord2f (newsl, newtl);
	glVertex2f (x, y);
	glTexCoord2f (newsh, newtl);
	glVertex2f (x+width, y);
	glTexCoord2f (newsh, newth);
	glVertex2f (x+width, y+height);
	glTexCoord2f (newsl, newth);
	glVertex2f (x, y+height);
	glEnd ();
}


/*
=============
Draw_TransPic
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{
	if (x < 0 || (unsigned)(x + pic->width) > vid.width || y < 0 ||
		 (unsigned)(y + pic->height) > vid.height)
	{
		Sys_Error ("Draw_TransPic: bad coordinates (%d, %d)", x, y);
	}
		
	Draw_Pic (x, y, pic);
}

/*
=============
Draw_TransPicTranslate

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
	int		size;
	int		i;
	byte	*dst;
	byte	*src;
	byte	*data;
	byte	*trans = NULL;
	int		transsize = 0;
	char	name[64];
	glpic_t	*gl;

	data = menuplyr_pixels;
	sprintf (name, "gfx/menuplyr.lmp");
	size = pic->width * pic->height;

	trans = (byte *)COM_AllocBuf ("Draw_TransPicTranslate", trans, &transsize, size, 64 * 64, name);

	dst = trans;
	src = data;

	for (i=0; i<size; i++)
		*dst++ = translation[*src++];

	data = trans;

	gl = (glpic_t *)pic->data;
	gl->gltexture = GL_LoadTexture (NULL, name, pic->width, pic->height, SRC_INDEXED, data, TEXPREF_ALPHA | TEXPREF_PAD | TEXPREF_NOPICMIP);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;

	Draw_Pic (x, y, pic);

	COM_FreeBuf (trans);
}


/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
	qpic_t *pic;
	int y;
	float conalpha;

	pic = Draw_CachePic ("gfx/conback.lmp");
	pic->width = vid.width;
	pic->height = vid.height;

	conalpha = (con_forcedup) ? 1.0 : CLAMP(0.0, r_conalpha.value, 1.0);

	y = (vid.height * 3) >> 2;

	if (lines > y)
		Draw_Pic(0, lines - vid.height, pic);
	else
		Draw_AlphaPic (0, lines - vid.height, pic, (float)(2 * conalpha * lines)/y);
}


/*
=============
Draw_TileClear

This repeats a 64*64 tile graphic to fill the screen around a sized down
refresh window.
=============
*/
void Draw_TileClear (int x, int y, int w, int h)
{
	glpic_t	*gl;

	gl = (glpic_t *)draw_backtile->data;

	glColor3f (1,1,1);
	GL_Bind (gl->gltexture);
	glBegin (GL_QUADS);
	glTexCoord2f (x/64.0, y/64.0);
	glVertex2f (x, y);
	glTexCoord2f ( (x+w)/64.0, y/64.0);
	glVertex2f (x+w, y);
	glTexCoord2f ( (x+w)/64.0, (y+h)/64.0);
	glVertex2f (x+w, y+h);
	glTexCoord2f ( x/64.0, (y+h)/64.0 );
	glVertex2f (x, y+h);
	glEnd ();
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	byte *pal = (byte *)d_8to24table; // use d_8to24table instead of host_basepal
	float alpha = 1;

	glDisable (GL_TEXTURE_2D);
	glEnable (GL_BLEND); // for alpha
	glDisable (GL_ALPHA_TEST); // for alpha
	glColor4f (pal[c*4]/255.0, pal[c*4+1]/255.0, pal[c*4+2]/255.0, alpha); // added alpha

	glBegin (GL_QUADS);
	glVertex2f (x,y);
	glVertex2f (x+w, y);
	glVertex2f (x+w, y+h);
	glVertex2f (x, y+h);
	glEnd ();

	glColor3f (1,1,1);
	glDisable (GL_BLEND); // for alpha
	glEnable (GL_ALPHA_TEST); // for alpha
	glEnable (GL_TEXTURE_2D);
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	glEnable (GL_BLEND);
	glDisable (GL_ALPHA_TEST);
	glDisable (GL_TEXTURE_2D);
	glColor4f (0, 0, 0, 0.5);

	glBegin (GL_QUADS);
	glVertex2f (0,0);
	glVertex2f (vid.width, 0);
	glVertex2f (vid.width, vid.height);
	glVertex2f (0, vid.height);
	glEnd ();

	glColor4f (1,1,1,1);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
	glDisable (GL_BLEND);

	Sbar_Changed();
}

//=============================================================================

/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.
================
*/
void Draw_BeginDisc (void)
{
	if (!draw_disc)
		return;
	glDrawBuffer  (GL_FRONT);
	Draw_Pic (vid.width - 24, 0, draw_disc);
	glDrawBuffer  (GL_BACK);
}


/*
================
Draw_EndDisc

Erases the disc icon.
Call after completing any disc IO
================
*/
void Draw_EndDisc (void)
{
}

/*
================
GL_Set2D

Setup as if the screen was 320*200
================
*/
void GL_Set2D (void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();

	glOrtho (0, glwidth, glheight, 0, -99999, 99999);
	glViewport (glx, gly, glwidth, glheight);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glDisable (GL_BLEND);
	glEnable (GL_ALPHA_TEST);

	glColor4f (1,1,1,1);
}

//====================================================================


/*
================
GL_ResampleTextureQuality
================
*/
static void GL_ResampleTextureQuality (unsigned *in, int inwidth, int inheight, unsigned *out,  int outwidth, int outheight, qboolean alpha)
{
	byte	 *nwpx, *nepx, *swpx, *sepx, *dest, *inlimit = (byte *)(in + inwidth * inheight);
	unsigned xfrac, yfrac, x, y, modx, mody, imodx, imody, injump, outjump;
	int	 i, j;

	// Make sure "out" size is at least 2x2!
	xfrac = ((inwidth-1) << 16) / (outwidth-1);
	yfrac = ((inheight-1) << 16) / (outheight-1);
	y = outjump = 0;

	for (i=0; i<outheight; i++)
	{
		mody = (y>>8) & 0xFF;
		imody = 256 - mody;
		injump = (y>>16) * inwidth;
		x = 0;

		for (j=0; j<outwidth; j++)
		{
			modx = (x>>8) & 0xFF;
			imodx = 256 - modx;

			nwpx = (byte *)(in + (x>>16) + injump);
			nepx = nwpx + sizeof(int);
			swpx = nwpx + inwidth * sizeof(int); // Next line

			// Don't exceed "in" size
			if (swpx + sizeof(int) >= inlimit)
			{
//				Con_Error ("GL_ResampleTextureQuality: %4d\n", swpx + sizeof(int) - inlimit);
				swpx = nwpx; // There's no next line
			}

			sepx = swpx + sizeof(int);

			dest = (byte *)(out + outjump + j);

			dest[0] = (nwpx[0]*imodx*imody + nepx[0]*modx*imody + swpx[0]*imodx*mody + sepx[0]*modx*mody)>>16;
			dest[1] = (nwpx[1]*imodx*imody + nepx[1]*modx*imody + swpx[1]*imodx*mody + sepx[1]*modx*mody)>>16;
			dest[2] = (nwpx[2]*imodx*imody + nepx[2]*modx*imody + swpx[2]*imodx*mody + sepx[2]*modx*mody)>>16;
			if (alpha)
				dest[3] = (nwpx[3]*imodx*imody + nepx[3]*modx*imody + swpx[3]*imodx*mody + sepx[3]*modx*mody)>>16;
			else
				dest[3] = 255;

			x += xfrac;
		}
		outjump += outwidth;
		y += yfrac;
	}
}

/*
================
GL_ResampleTexture
================
*/
void GL_ResampleTexture (unsigned *in, int inwidth, int inheight, unsigned *out,  int outwidth, int outheight, qboolean alpha)
{
	int		i, j;
	unsigned	*inrow;
	unsigned	frac, fracstep;

	// Sanity ...
	if (inwidth <= 0 || inheight <= 0 || outwidth <= 0 || outheight <= 0 ||
		inwidth * 0x10000 & 0xC0000000 || inheight * outheight & 0xC0000000 ||
		inwidth * inheight & 0xC0000000)
		Sys_Error ("GL_ResampleTexture: invalid parameters (in:%dx%d, out:%dx%d)", inwidth, inheight, outwidth, outheight);

	if (gl_texturequality.value)
	{
		GL_ResampleTextureQuality (in, inwidth, inheight, out, outwidth, outheight, alpha);
		return;
	}

	fracstep = inwidth*0x10000/outwidth;
	for (i=0 ; i<outheight ; i++, out += outwidth)
	{
		inrow = in + inwidth*(i*inheight/outheight);
		frac = fracstep >> 1;
		for (j=0 ; j<outwidth ; j+=4)
		{
			out[j] = inrow[frac>>16];
			frac += fracstep;
			out[j+1] = inrow[frac>>16];
			frac += fracstep;

			// Don't exceed "in" size
			if ((int)(frac>>16) >= inwidth)
				break;

			out[j+2] = inrow[frac>>16];
			frac += fracstep;
			out[j+3] = inrow[frac>>16];
			frac += fracstep;
		}
	}
}

/*
================
MipMap

Operates in place, quartering the size of the texture
================
*/
static void MipMap (byte *in, int width, int height)
{
	int		i, j;
	byte	*out;

	width <<=2;
	height >>= 1;
	out = in;
	for (i=0 ; i<height ; i++, in+=width)
	{
		for (j=0 ; j<width ; j+=8, out+=4, in+=8)
		{
			out[0] = (in[0] + in[4] + in[width+0] + in[width+4])>>2;
			out[1] = (in[1] + in[5] + in[width+1] + in[width+5])>>2;
			out[2] = (in[2] + in[6] + in[width+2] + in[width+6])>>2;
			out[3] = (in[3] + in[7] + in[width+3] + in[width+7])>>2;
		}
	}
}

/*
================
GL_MipMap
================
*/
void GL_MipMap (unsigned *scaled, int *scaled_width, int *scaled_height)
{
	MipMap ((byte *)scaled, *scaled_width, *scaled_height);

	*scaled_width >>= 1;
	*scaled_height >>= 1;
	if (*scaled_width < 1)
		*scaled_width = 1;
	if (*scaled_height < 1)
		*scaled_height = 1;
}

/*
================
ScaleSize
================
*/
static int ScaleSize (int OldSize, qboolean Force)
{
	int NewSize, NextSize;

	// gl_texturequality == 0 : GLQuake compatibility
	// gl_texturequality == 1 : default, better resampling, less blurring of non-mipmapped texes
	// gl_texturequality == 2 : better resampling, less blurring of all texes, requires a lot of memory
	if (gl_texturequality.value < 1 || Force)
		NextSize = OldSize;
	else
		NextSize = 3 * OldSize / 2; // Avoid unfortunate resampling

	for (NewSize = 1; NewSize < NextSize && NewSize != OldSize; NewSize <<= 1)
		;

	return NewSize;
}

/*
===============
GL_AlphaEdgeFix

eliminate pink edges on sprites, etc.
operates in place on 32bit data
===============
*/
static void GL_AlphaEdgeFix (byte *data, int width, int height)
{
	int i,j,n=0,b,c[3]={0,0,0},lastrow,thisrow,nextrow,lastpix,thispix,nextpix;
	byte *dest = data;

	for (i=0; i<height; i++)
	{
		lastrow = width * 4 * ((i == 0) ? height-1 : i-1);
		thisrow = width * 4 * i;
		nextrow = width * 4 * ((i == height-1) ? 0 : i+1);

		for (j=0; j<width; j++, dest+=4)
		{
			if (dest[3]) //not transparent
				continue;

			lastpix = 4 * ((j == 0) ? width-1 : j-1);
			thispix = 4 * j;
			nextpix = 4 * ((j == width-1) ? 0 : j+1);

			b = lastrow + lastpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = thisrow + lastpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = nextrow + lastpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = lastrow + thispix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = nextrow + thispix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = lastrow + nextpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = thisrow + nextpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = nextrow + nextpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}

			//average all non-transparent neighbors
			if (n)
			{
				dest[0] = (byte)(c[0]/n);
				dest[1] = (byte)(c[1]/n);
				dest[2] = (byte)(c[2]/n);

				n = c[0] = c[1] = c[2] = 0;
			}
		}
	}
}

/*
===============
GL_Upload32 -- handles 32bit source data
===============
*/
void GL_Upload32 (gltexture_t *glt, unsigned *data)
{
	int			internalformat;
	int			scaled_width, scaled_height;
	int			picmip;
	static unsigned	*scaled = NULL;
	static int	scaledsize = 0;

	scaled_width = ScaleSize (glt->width, gl_texturequality.value == 1 && (glt->flags & TEXPREF_MIPMAP));
	scaled_height = ScaleSize (glt->height, gl_texturequality.value == 1 && (glt->flags & TEXPREF_MIPMAP));

	if (glt->width && glt->height)
	{
		// Preserve proportions
		while (glt->width > glt->height && scaled_width < scaled_height)
			scaled_width *= 2;

		while (glt->width < glt->height && scaled_width > scaled_height)
			scaled_height *= 2;
	}

	// Note: Can't use Con_Printf here!
	if (developer.value > 1 && (scaled_width != ScaleSize (glt->width, true) || scaled_height != ScaleSize (glt->height, true)))
		Con_DPrintf ("GL_Upload32: in:%dx%d, out:%dx%d, '%s'\n", glt->width, glt->height, scaled_width, scaled_height, glt->name);

	// Prevent too large or too small images (might otherwise crash resampling)
	scaled_width = CLAMP(2, scaled_width, gl_texturemaxsize);
	scaled_height = CLAMP(2, scaled_height, gl_texturemaxsize);

	// Allocate
	scaled = (unsigned *)COM_AllocBuf ("GL_Upload32", scaled, &scaledsize, scaled_width * scaled_height * sizeof(unsigned), 0, glt->name);

	// Resample up
	if (glt->width && glt->height) // Don't resample 0-sized images
		GL_ResampleTexture (data, glt->width, glt->height, scaled, scaled_width, scaled_height, (glt->flags & TEXPREF_ALPHA));
	else
		memcpy (scaled, data, scaled_width * scaled_height * rgba_bytes); // FIXME: 0-sized texture, so we just copy it

	// mipmap down
	picmip = (glt->flags & TEXPREF_NOPICMIP) ? 0 : max((int)gl_picmip.value, 0);
	if (glt->flags & TEXPREF_MIPMAP)
	{
		int i;

		// Only affect mipmapped texes, typically not console graphics
		for (i = 0; i < picmip && (scaled_width > 1 || scaled_height > 1); ++i)
		{
			GL_MipMap (scaled, &scaled_width, &scaled_height);
			if (glt->flags & TEXPREF_ALPHA)
				GL_AlphaEdgeFix ((byte *)scaled, scaled_width, scaled_height);
		}
	}

	// upload
	GL_Bind (glt);
	internalformat = (glt->flags & TEXPREF_ALPHA) ? gl_alpha_format : gl_solid_format;
	glTexImage2D (GL_TEXTURE_2D, 0, internalformat, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);

	// upload mipmaps
	if (glt->flags & TEXPREF_MIPMAP)
	{
		int miplevel;

		miplevel = 0;
		while (scaled_width > 1 || scaled_height > 1)
		{
			GL_MipMap (scaled, &scaled_width, &scaled_height);
			miplevel++;
			glTexImage2D (GL_TEXTURE_2D, miplevel, internalformat, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
		}
	}

	// set filter modes
	GL_SetFilterModes (glt);
}

/*
================
GL_UploadLuma -- handles bloom data
================
*/
void GL_UploadLuma (gltexture_t *glt, unsigned *data)
{
	// upload it
	GL_Bind (glt);
	glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, glt->width, glt->height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

	// set filter modes
	GL_SetFilterModes (glt);
}

/*
================
GL_UploadLightmap -- handles lightmap data
================
*/
void GL_UploadLightmap (gltexture_t *glt, byte *data)
{
	// upload it
	GL_Bind (glt);
	glTexImage2D (GL_TEXTURE_2D, 0, gl_lightmap_format, glt->width, glt->height, 0, gl_lightmap_format, GL_UNSIGNED_BYTE, data);

	// set filter modes
	GL_SetFilterModes (glt);
}

/*
===============
GL_Upload8 -- handles 8bit source data, then passes it to GL_Upload32
===============
*/
void GL_Upload8 (gltexture_t *glt, byte *data)
{
	int			i, size;
	int			p;
	static unsigned		*trans = NULL;
	static int			transsize = 0;
	int					maxsize;

	// HACK HACK HACK -- taken from fitzquake
	if (strstr(glt->name, "shot1sid") && glt->width==32 && glt->height==32 && CRC_Block(data, 1024) == 65393)
	{
		// This texture in b_shell1.bsp has some of the first 32 pixels painted white.
		// They are invisible in software, but look really ugly in GL. So we just copy
		// 32 pixels from the bottom to make it look nice.
		memcpy (data, data + 32*31, 32);
	}

	maxsize = gl_texturemaxsize * gl_texturemaxsize;
	size = glt->width * glt->height;

	if (size & 3)
	{
		Con_Warning ("GL_Upload8: size %d is not a multiple of 4 in '%s'\n", size, glt->name);
	}

	// Allocate
	trans = (unsigned *)COM_AllocBuf ("GL_Upload8", trans, &transsize, size * sizeof(unsigned), maxsize * sizeof(unsigned), glt->name);

	// detect false alpha cases
	if (glt->flags & TEXPREF_ALPHA && !(glt->flags & TEXPREF_CONCHARS))
	{
		for (i=0 ; i<size ; i++)
		{
			p = data[i];
			if (p == 255) //transparent index
				break;
		}
		if (i == size)
			glt->flags -= TEXPREF_ALPHA;
	}

	// choose palette and convert to 32bit
	if (glt->flags & TEXPREF_FULLBRIGHT)
	{
		for (i=0 ; i<size ; ++i)
		{
			p = data[i];
			trans[i] = d_8to24table_fbright[p];
		}
	}
	else if (glt->flags & TEXPREF_NOBRIGHT)
	{
		for (i=0 ; i<size ; ++i)
		{
			p = data[i];
			trans[i] = d_8to24table_nobright[p];
		}
	}
	else if (glt->flags & TEXPREF_CONCHARS)
	{
		for (i=0 ; i<size ; ++i)
		{
			p = data[i];
			trans[i] = d_8to24table_conchars[p];
		}
	}
	else
	{
		for (i=0 ; i<size ; ++i)
		{
			p = data[i];
			trans[i] = d_8to24table[p];
		}
	}

	// fix edges
	if (glt->flags & TEXPREF_ALPHA)
		GL_AlphaEdgeFix ((byte *)trans, glt->width, glt->height);

	// upload it
	GL_Upload32 (glt, trans);
}


/*
================
GL_FindTexture
================
*/
gltexture_t *GL_FindTexture (model_t *owner, char *name)
{
	gltexture_t	*glt;

	if (name)
	{
		for (glt = active_gltextures; glt; glt = glt->next)
			if (glt->owner == owner && !strcmp (glt->name, name))
				return glt;
	}

	return NULL;
}


/*
================
GL_NewTexture
================
*/
gltexture_t *GL_NewTexture (void)
{
	gltexture_t *glt;

	if (numgltextures >= MAX_GLTEXTURES)
		Sys_Error ("GL_NewTexture: cache full, max is %i textures", MAX_GLTEXTURES);

	glt = free_gltextures;
	free_gltextures = glt->next;
	glt->next = active_gltextures;
	active_gltextures = glt;

	glGenTextures(1, (GLuint *)&glt->texnum);
	numgltextures++;
	return glt;
}


/*
================
GL_FreeTexture
================
*/
void GL_FreeTexture (gltexture_t *free)
{
	gltexture_t *glt;

	if (free == NULL)
	{
		Con_SafePrintf ("GL_FreeTexture: NULL texture\n");
		return;
	}

	if (active_gltextures == free)
	{
		active_gltextures = free->next;
		free->next = free_gltextures;
		free_gltextures = free;

		glDeleteTextures(1, (const GLuint *)&free->texnum);
		numgltextures--;
		return;
	}

	for (glt = active_gltextures; glt; glt = glt->next)
	{
		if (glt->next == free)
		{
			glt->next = free->next;
			free->next = free_gltextures;
			free_gltextures = free;

			glDeleteTextures(1, (const GLuint *)&free->texnum);
			numgltextures--;
			return;
		}
	}

	Con_SafePrintf ("GL_FreeTexture: not found\n");
}

/*
================
GL_FreeTextures
================
*/
void GL_FreeTextures (model_t *owner)
{
	gltexture_t *glt, *next;

	for (glt = active_gltextures; glt; glt = next)
	{
		next = glt->next;
		if (glt && glt->owner == owner)
			GL_FreeTexture (glt);
	}
}

/*
================
GL_LoadTexture
================
*/
gltexture_t *GL_LoadTexture (model_t *owner, char *name, int width, int height, enum srcformat format, byte *data, unsigned flags)
{
	int		size = 0; // keep compiler happy
	gltexture_t	*glt;
	unsigned short crc;

	if (isDedicated)
		return NULL; // No textures in dedicated mode

	// cache check
	switch (format)
	{
		case SRC_INDEXED:
			size = width * height * indexed_bytes;
			break;
		case SRC_LIGHTMAP:
			size = width * height * lightmap_bytes;
			break;
		case SRC_RGBA:
			size = width * height * rgba_bytes;
			break;
		case SRC_BLOOM:
			size = width * height * rgba_bytes;
			break;
	}

	if (size == 0)
		Con_DPrintf ("GL_LoadTexture: texture '%s' has size 0\n", name);

	// Sanity check, max = 32kx32k
	if (width < 0 || height < 0 || size > 0x40000000)
		Sys_Error ("GL_LoadTexture: texture '%s' has invalid size (%dM, max = %dM)", name, size / (1024 * 1024), 0x40000000 / (1024 * 1024));

	crc = CRC_Block(data, size);

	if ((flags & TEXPREF_OVERWRITE) && (glt = GL_FindTexture (owner, name)))
	{
		if (glt->source_crc == crc)
			return glt;
	}
	else
		glt = GL_NewTexture ();

	// copy data
	glt->owner = owner;
	strncpy (glt->name, name, sizeof(glt->name));
	glt->width = width;
	glt->height = height;
	glt->flags = flags;
	glt->source_format = format;
	glt->source_width = width;
	glt->source_height = height;
	glt->source_crc = crc;

	switch (glt->source_format)
	{
		case SRC_INDEXED:
			GL_Upload8 (glt, data);
			break;
		case SRC_LIGHTMAP:
			GL_UploadLightmap (glt, data);
			break;
		case SRC_RGBA:
			GL_Upload32 (glt, (unsigned *)data);
			break;
		case SRC_BLOOM:
			GL_UploadLuma (glt, (unsigned *)data);
			break;
	}

	return glt;
}

